from ontic import (core,
                   property,
                   schema,
                   type)
from ontic.property import OnticProperty
from ontic.schema import Schema
from ontic.type import OnticType

__all__ = [
    'core',
    'property',
    'schema',
    'type',
    'OnticProperty',
    'Schema',
    'OnticType',
]



"""Module containing definition for schemata type and base validation functions.

.. image:: images/ontic_meta.jpg

.. contents::

"""
import re
from datetime import date, datetime, time
from typing import Any, NoReturn, Callable

import ontic
from ontic import core
from ontic.schema import Schema


class Meta(core.Core):
    """Interface for type definition of **Ontic** schema defined classes.

    Dict Style Initialization
        Meta() -> new empty Meta

        Meta(mapping) -> new Meta initialized from a mapping object's
            (key, value) pairs

        Meta(iterable) -> new Meta initialized as if via::

            d = Meta()
            for k, v in iterable:
                d[k] = v

        Meta(**kwargs) -> new Meta initialized with the name=value
        pairs in the keyword argument list.  For example::

            Meta(one=1, two=2)
    """
    ONTIC_SCHEMA = core.Core()

    @classmethod
    def get_schema(cls) -> Schema:
        """Returns the schema object for the a given type definition.

        :return: The schema metadata definition for a :class:`PropertySchema`
            or a :class:`ontic.OnticType` derived child class.
        :rtype: :class:`ontic.Schema`
        """
        return cls.ONTIC_SCHEMA


#: The set of supported collection types.
COLLECTION_TYPES = {dict, list, set, tuple}

#: The set of types that can be compared with inequality operators.
COMPARABLE_TYPES = {complex, date, datetime, float, int, time, tuple}

#: The set of types that may be limited in size.
BOUNDABLE_TYPES = {str, list, dict, set}

#: Used to convert the string declaration of attribute type to native type.
TYPE_MAP = {
    'bool': bool,
    bool: bool,
    'complex': complex,
    complex: complex,
    'date': date,
    date: date,
    'datetime': datetime,
    datetime: datetime,
    'dict': dict,
    dict: dict,
    'float': float,
    float: float,
    'int': int,
    int: int,
    'list': list,
    list: list,
    'None': None,
    None: None,
    Meta: Meta,
    'set': set,
    set: set,
    'str': str,
    str: str,
    'time': time,
    time: time,
    'tuple': tuple,
    tuple: tuple,
}

TYPE_SET = (
    bool,
    complex,
    date,
    datetime,
    dict,
    float,
    int,
    list,
    set,
    str,
    time,
    tuple,
)


def validate_value(property_schema: 'OnticProperty', value: Any) -> list[str]:
    """Method to validate a given value against a given property schema.

    :param property_schema: The property schema that contains the validation
        rules.
    :param value: The value that is to be validated.
    :return: A list that is utilized to collect the errors found
        during schema validation.
    """
    value_errors = []
    # required: True | False
    if property_schema.required and value is None:
        value_errors.append('The value for "%s" is required.' %
                            property_schema.name)
        return value_errors  # No other validation can occur without a value

    if value is not None:
        validate_non_none_value(property_schema, value, value_errors)

    return value_errors


def validate_non_none_value(
        property_schema: 'OnticProperty',
        value: Any,
        value_errors: list[str]) -> NoReturn:
    """Validates an **Ontic** object value that is not None.

    This method validates singular and collection values. This method
    does not perform *Required* validation, as it is assumed that the
    value is not None.

    :param property_schema: The property schema to utilize for validation.
    :type property_schema: :class:`ontic.OnticProperty`
    :param value: The non-None value to be validated.
    :type value: object
    :param value_errors: A list of errors found for a given value. If any
        given validator method fails, it will append it error message to
        the value_errors list.
    :type value_errors: list<str>
    :rtype: None
    """
    if not property_schema.type:
        # if no schema_type, then just check that
        # the value is in an enum if necessary.
        if not enum_validation(property_schema, value):
            value_errors.append(
                'The value "%s" for "%s" not in enumeration %s.' %
                (value, property_schema.name, _generate_sorted_list(
                    property_schema.enum)))
            return  # No further processing can occur.
    else:
        # type checking
        if not isinstance(value, property_schema.type):
            value_errors.append(
                'The value for "%s" is not of type "%s": %s' %
                (property_schema.name, property_schema.type, str(value)))
            # If not of the expected type, than can't further
            # validate without errors.
            return

        if property_schema.type in COLLECTION_TYPES:
            validate_collection_members(
                property_schema, value, value_errors)
        else:
            non_none_singular_validation(
                property_schema, value, value_errors)


def validate_collection_members(
        property_schema: 'OnticProperty',
        value: Any,
        value_errors: list[str]) -> NoReturn:
    """Method to validate the members of a collection.

    This method only operates on *list* and *set* collection types.

    :param property_schema: The property schema to utilize for validation.
    :type property_schema: :class:`ontic.OnticProperty`
    :param value: The collection whose members will be validated.
    :type value: list, set
    :param value_errors: A list of errors found for a given collection.
        If any members fail validation, the error condition will be
        listed in value_errors list.
    :type value_errors: list<str>
    :rtype: None
    """
    if not min_validation(property_schema, value):
        value_errors.append('The value of "%s" for "%s" fails min of %s.' %
                            (value, property_schema.name, property_schema.min))

    if not max_validation(property_schema, value):
        value_errors.append('The value of "%s" for "%s" fails max of %s.' %
                            (value, property_schema.name, property_schema.max))

    if property_schema.type in {list, set}:
        validators = list()

        if property_schema.enum:
            validators.append(validate_member_enum)
        if property_schema.member_type:
            validators.append(validate_member_type)
        if property_schema.regex and property_schema.member_type == str:
            validators.append(validate_member_regex)
        if property_schema.member_min:
            validators.append(validate_member_min)
        if property_schema.member_max:
            validators.append(validate_member_max)

        for member_value in value:
            execute_collection_validators(
                member_value,
                property_schema,
                validators,
                value_errors)


def execute_collection_validators(
        member_value: [str, int, float, date, datetime, time],
        property_schema: 'OnticProperty',
        validators: list[Callable],
        value_errors: list[str]) -> NoReturn:
    """Method to execute a list of validators on a given collection.

    :param member_value: The member of the collection property to validate.
    :param property_schema: The property schema to utilize for validation.
    :param validators: A list of validation methods to execute.
    :param value_errors: A list of errors found for a given value. If any
        given validator method fails, it will append it error message to
        the value_errors list.
    """
    for validator in validators:
        validator(member_value, property_schema, value_errors)


def validate_member_enum(
        member_value: [str, int, float, date, datetime, time],
        property_schema: 'OnticProperty',
        value_errors: list[str]) -> NoReturn:
    """Validate a member of a collection is within a defined enumeration.

    :param member_value: The member of the collection property to
        validate.
    :param property_schema: The property schema to utilize for validation.
    :param value_errors: A list of errors found for a given value. If the
        validate fails, then an error message is added to the
        value_errors list.
    """
    if not enum_validation(property_schema, member_value):
        value_errors.append(
            'The value "%s" for "%s" not in enumeration %s.' %
            (member_value, property_schema.name, _generate_sorted_list(
                property_schema.enum)))


def validate_member_type(
        member_value: [str, int, float, date, datetime, time],
        property_schema: 'OnticProperty',
        value_errors: list[str]) -> NoReturn:
    """Validate a member of a collection is of a given type.

    :param member_value: The member value of the collection property to
        validate.
    :type member_value: object
    :param property_schema: The property schema to utilize for validation.
    :type property_schema: :class:`ontic.OnticProperty`
    :param value_errors: A list of errors found for a given value. If the
        validation fails, then an error message is added to the
        value_errors list.
    :type value_errors: list<str>
    :rtype: None
    """
    if not isinstance(member_value, property_schema.member_type):
        value_errors.append(
            'The value "%s" for "%s" is not of type "%s".' %
            (str(member_value), property_schema.name,
             property_schema.member_type))


def validate_member_regex(
        member_value: [str, int, float, date, datetime, time],
        property_schema: 'OnticProperty',
        value_errors: list[str]) -> NoReturn:
    """Validate a member of a collection against a defined regex.

    :param member_value: The member value of the collection property to
        validate.
    :param property_schema: The property schema to utilize for validation.
    :param value_errors: A list of errors found for a given value. If the
        validation fails, then an error message is added to the
        value_errors list.
    """
    if not re.match(property_schema.regex, member_value):
        value_errors.append(
            'Value "%s" for "%s" does not meet regex: %s' %
            (member_value, property_schema.name, property_schema.regex))


def validate_member_min(
        member_value: [str, int, float, date, datetime, time],
        property_schema: 'OnticProperty',
        value_errors: list[str]) -> NoReturn:
    """Validate a member of a collection for minimum allowable value.

    :param member_value: The member value of the collection property to
        validate.
    :param property_schema: The property schema to utilize for validation.
    :param value_errors: A list of errors found for a given value. If the
        validation fails, then an error message is added to the
        value_errors list.
    """
    if property_schema.member_type is str:
        if len(member_value) < property_schema.member_min:
            value_errors.append(
                'The value of "%s" for "%s" fails min length of %s.' %
                (member_value, property_schema.name,
                 property_schema.member_min))

    if property_schema.member_type in COMPARABLE_TYPES:
        if member_value < property_schema.member_min:
            value_errors.append(
                'The value of "%s" for "%s" fails min size of %s.' %
                (member_value, property_schema.name,
                 property_schema.member_min))


def validate_member_max(member_value: [str, int, float, date, datetime, time],
                        property_schema: 'OnticProperty',
                        value_errors: list[str]) -> NoReturn:
    """Validate a member of a collection for maximum allowable value.

    :param member_value: The member value of the collection property to
        validate.
    :param property_schema: The property schema to utilize for validation.
    :param value_errors: A list of errors found for a given value. If the
        validation fails, then an error message is added to the
        value_errors list.
    """
    if property_schema.member_type is str:
        if len(member_value) > property_schema.member_max:
            value_errors.append(
                'The value of "%s" for "%s" fails max length of %s.' %
                (member_value, property_schema.name,
                 property_schema.member_max))

    if property_schema.member_type in COMPARABLE_TYPES:
        if member_value > property_schema.member_max:
            value_errors.append(
                'The value of "%s" for "%s" fails max size of %s.' %
                (member_value, property_schema.name,
                 property_schema.member_max))


def enum_validation(
        property_schema: 'OnticProperty',
        value: [str, int, float, date, datetime, time]) -> bool:
    """Validate a non-collection property for value in an enumeration set.

    :param property_schema: The property schema to utilize for validation.
    :param value: The value of the property to be validated.
    :return: True if the validation is successful, else False.
    """
    if property_schema.enum:
        if value not in property_schema.enum:
            return False
    return True


def min_validation(
        property_schema: 'OnticProperty',
        value: [str, int, float, date, datetime, time]) -> bool:
    """Validate a non-collection property for minimum allowable value.

    :param property_schema: The property schema to utilize for validation.
    :param value: The value of the property to be validated.
    :return: True if the validation is successful, else False.
    """
    if property_schema.min:
        if property_schema.type in BOUNDABLE_TYPES:
            if len(value) < property_schema.min:
                return False
        if property_schema.type in COMPARABLE_TYPES:
            if value < property_schema.min:
                return False

    return True


def max_validation(
        property_schema: 'OnticProperty',
        value: [str, int, float, date, datetime, time]) -> bool:
    """Validates a non-collection property for maximum allowable value.

    :param property_schema: The property schema to utilize for validation.
    :param value: The value of the property to be validated.
    :return: True if the validation is successful, else False.
    """
    if property_schema.max:
        if property_schema.type in BOUNDABLE_TYPES:
            if len(value) > property_schema.max:
                return False
        if property_schema.type in COMPARABLE_TYPES:
            if value > property_schema.max:
                return False

    return True


def non_none_singular_validation(
        property_schema: 'OnticProperty',
        value: [str, int, float, date, datetime, time],
        value_errors: list[str]) -> NoReturn:
    """Method to validate an object value meets schema requirements.

    This method validates non-collection properties. The method should
    only be used for non-None values.

    :param property_schema: The schema definition for the target property.
    :param value: The value to be tested against the given schema.
    :param value_errors: A list of the validation errors discovered. The
        value errors will be added to if the given value fails validation.
    """
    # enum
    if not enum_validation(property_schema, value):
        value_errors.append('The value "%s" for "%s" not in enumeration %s.' %
                            (value, property_schema.name,
                             _generate_sorted_list(property_schema.enum)))

    # min
    if not min_validation(property_schema, value):
        value_errors.append('The value of "%s" for "%s" fails min of %s.' %
                            (value, property_schema.name, property_schema.min))

    # max
    if not max_validation(property_schema, value):
        value_errors.append('The value of "%s" for "%s" fails max of %s.' %
                            (value, property_schema.name, property_schema.max))

    # regex validation
    if property_schema.regex:
        if property_schema.type is str and value != '':
            if not re.match(property_schema.regex, value):
                value_errors.append(
                    'Value "%s" for %s does not meet regex: %s' %
                    (value, property_schema.name, property_schema.regex))


def _generate_sorted_list(some_collection: list[Any]) -> list[Any]:
    """Attempt to generate a sorted list from a collection.

    :param some_collection: A collection that will attempt to be sorted.
    :return: The sorted collection if possible, else return original collection.
    """
    return sorted(some_collection, key=lambda x: (x is None, str(x)))



"""An exception to signal a validation failure."""


class ValidationException(Exception):
    """Exception used to signal validation errors."""

    def __init__(self, validation_errors=None):
        """

        :param validation_errors: A list of validation failures that are
            triggering the *ValidationException*.
        :type validation_errors: list<str>
        """
        self._validation_errors = validation_errors if validation_errors else []
        message = str.join(' \n', validation_errors)
        super(ValidationException, self).__init__(message)

    @property
    def validation_errors(self) -> list[str]:
        """List of validation errors that triggered the *ValidationException*.

        :return: A list of the validation errors encountered.
        """
        return self._validation_errors



"""The classes utilized to construct schemas for object definitions.

.. image:: images/ontic_schema.jpg

.. contents::

======
Usage
======

The *schema_type* module contains the class :class:`Schema` and a set of
functions to create and validate schema. *Schema* are used to validate
:class:`ontic.ontic_type.OnticType` derived instances.

Creating Schema
----------------

A *Schema* is defined as a dictionary with the key entry being the name
of the property. The value portion of the dictionary is a
:class:`ontic.meta_type.PropertySchema` instance::

    >>> a_schema = Schema([
    ...     property.OnticProperty(name='property_name', type= 'str')
    ... ])

While the example above give a strict definition of a schema, creation of a
schema can omit the use of the *PropertySchema*, as the *Schema*
instantiation will convert a dict value to a *PropertySchema* object. The
above example can be simplified to be::

    >>> a_schema = Schema([
    ...     {'name': 'property_name', 'type': 'str'}
    ... ])

The *Schema* also supports the dict style of instantiation via parameter
naming::

    >>> a_schema = Schema(property_name={'type': 'str'})

Dynamic Schema
-----------------

In cases where necessary, a *Schema* can be created dynamically::

    >>> a_schema = Schema()
    >>> a_schema.add(property.OnticProperty({'name': 'a', 'type': 'str'}))

To aid in the handling of dynamic models, utilize the :meth:`perfect_schema`
and :meth:`validate_schema`.

    >>> perfect_schema(a_schema)
    >>> errors = validate_schema(a_schema)

"""
import logging
from typing import NoReturn

from ontic import core
from ontic import property
from ontic.validation_exception import ValidationException


class Schema(core.Core):
    """The type definition for a schema object.

    The **Schema** contains a dictionary of property field names and
    the corresponding **OnticProperty** definition.

    Example Schema representation::

        Schema({
          'some_property': Schema({
                'type': 'str',
                'required': True
            })
        })

    For a complete list of :class:`OnticProperty` settings, see
    :ref:`property-schema-settings-table`.
    """

    def __init__(self, *args, **kwargs):

        # Detect if the list is a list of dicts or OnticProperty types.
        schema_list = None
        if len(args) == 1 and isinstance(args[0], list):
            if len(args[0]) > 0 and isinstance(args[0][0], dict):
                schema_list = args[0]

        if schema_list:
            super(Schema, self).__init__()
            for some_property in schema_list:
                self.add(some_property)
            return  # Initialization completed for a list.

        super(Schema, self).__init__(*args, **kwargs)

        for key, value in self.items():
            if not isinstance(value, property.OnticProperty):
                try:
                    if isinstance(value, dict):
                        value['name'] = key  # Set name for consistency.
                    self[key] = property.OnticProperty(value)
                except Exception:
                    logging.exception(
                        'Exception while converting "%s" to OnticProperty', key)
                    raise

    def add(self, property_type: 'OnticProperty') -> NoReturn:
        """ Add a property definition to a schema.

        :param property_type:
        """
        if not isinstance(property_type, (dict, property.OnticProperty)):
            raise ValueError(
                '"property_type" must be dict or OnticProperty type.')
        if isinstance(property_type, property.OnticProperty):
            self[property_type.name] = property_type

        else:
            ontic_prop = property.OnticProperty(property_type)
            self[ontic_prop.name] = ontic_prop

    def perfect(self) -> NoReturn:
        """Method to clean and perfect a given schema.

        The *perfect* will fill in any missing schema settings for each of
        the :class:`OnticProperty`. This function should be used to ensure
        property schema completeness.
        """
        perfect_schema(self)

    def validate(self,
                 raise_validation_exception: bool = True) -> (None, list[str]):
        """Validate a given :class:`Schema`.

        This method will iterate through all of the
        :class:`OnticProperty` and validate that each definition is valid.
        The method will collect all of the errors and return those as a list
        of strings or raise a
        :class:`ontic.validation_exception.ValidationException`. The switch in
        behavior is determined by the *raise_validation_exception*

        :param raise_validation_exception: If True, then *validate_schema* will
            throw a *ValidationException* upon validation failure. If False,
            then a list of validation errors is returned. Defaults to True.
        :type raise_validation_exception: bool
        :return: List of errors found. Empty of no errors found.
        :rtype: list<str>
        :raises ValueError: *candidate_schema* is None, or not of type
            :class:`Schema`.
        :raises ValidationException: A property of *candidate_schema* does not
            meet schema requirements.
        """
        return validate_schema(self, raise_validation_exception)


def perfect_schema(ontic_schema: 'OnticProperty') -> NoReturn:
    """Method to clean and perfect a given schema.

    The *perfect_schema* will fill in any missing schema setting for each of
    the :class:`OnticProperty`. This function should be used to ensure
    property schema completeness.

    :param ontic_schema: The schema that is to be perfected.
    """
    if ontic_schema is None:
        raise ValueError('"ontic_schema" must be provided.')
    if not isinstance(ontic_schema, Schema):
        raise ValueError('"ontic_schema" argument must be of Schema type.')

    [property_schema.perfect() for property_schema in ontic_schema.values()]


def validate_schema(
        ontic_schema: 'OnticProperty',
        raise_validation_exception: bool = True) -> (None, list[str]):
    """Validate a given :class:`Schema`.

    This method will iterate through all of the :class:`OnticProperty` and
    validate that each definition is valid.  The method will collect all of
    the errors and return those as a list of strings or raise a
    :class:`ontic.validation_exception.ValidationException`. The switch in
    behavior is determined by the *raise_validation_exception*

    :param ontic_schema: The schema to be validated.
    :param raise_validation_exception: If True, then *validate_schema* will
        throw a *ValidationException* upon validation failure. If False, then a
        list of validation errors is returned. Defaults to True.
    :return: List of errors found. Empty of no errors found.
    :raises ValueError: *ontic_schema* is None, or not of type
        :class:`Schema`.
    :raises ValidationException: A property of *ontic_schema* does not
        meet schema requirements.
    """
    if ontic_schema is None:
        raise ValueError('"ontic_schema" argument must be provided.')
    if not isinstance(ontic_schema, Schema):
        raise ValueError('"ontic_schema" argument must be of Schema type.')

    value_errors = []
    for prop in ontic_schema.values():
        value_errors.extend(
            prop.validate(raise_validation_exception=False))

    if value_errors and raise_validation_exception:
        raise ValidationException(value_errors)

    return value_errors



"""The base class for deriving types that require schema support.


.. image:: images/ontic_property.jpg

.. contents::

======
Usage
======

The **meta_type** module contains three classes that are used in the
definition and instantiation of object instances.  The root class is
:class:`CoreType`, which is derived from the Python :class:`dict`. The
primary feature of the :class:`CoreType` is provide the means to access
properties of an object as a *dict* key or as object attribute.

Derived from :class:`CoreType` is :class:`MetaType`, which provides the
interface for retrieving a schema from an object schema definition via the
:meth:`MetaType.get_schema()`.

The :class:`PropertySchema` is utilized to define a schema for a single
property. This includes setting data type, required, and other such common
schema definition conventions. See :ref:`property-schema-settings-table` for
details on the :class:`PropertySchema` settings available.

It would be a rare case that requires the use of :class:`CoreType` or
:class:`MetaType`. For the majority of cases, the use of
:class:`PropertySchema` is sufficient. With that in mind, the remainder of
this section will focus on the use of :class:`PropertySchema`

Creating Property Schema
-------------------------

For general purposes, consider utilizing :class:`ontic.schema_type.SchemaType`,
for defining complete models. However, if you need validators for individual
properties, then direct use of :class:`PropertySchema` is a solution.

There are a number of ways to create a :class:`PropertySchema`. Take a look
at :class:`PropertySchema` class documentation for a complete exposition on
the means of instantiating an instance.

The most straight forward way to create an instance of a
:class:`PropertySchema`:

>>> prop_schema = OnticProperty(name='x', type='str', required=True, min=3)
>>> assert prop_schema == {'regex': None, 'enum': None, 'min': 3, 'default': None, 'max': None,
...     'required': True, 'member_type': None, 'member_min': None,
...     'type': str, 'member_max': None, 'name': 'x'}

Demonstrated above is the creation of a property schema of type string. In
addition the property schema forces the value of the property to required and
of minimum length of 3.

Along with the schema settings explicitly set in the constructor, there are a
number of other property schema settings that may be utilized. These
additional settings can be viewed in the output of the *prop_schema* object.
For the details on the property schema settings, see
:ref:`property-schema-settings-table`.

A :class:`PropertySchema` can be created or modified dynamically. If done so,
then the final schema instance should be validated with the use of the method
:meth:`validate_property_schema`.

Utilizing Property Schema
--------------------------

Validation of a value utilizing the *prop_schema* created, is done with the
:meth:`validate_value` method.

>>> prop_schema = OnticProperty(name='some_prop', type='str', required=True)
>>> some_value = 'The cat is on the roof.'
>>> prop_schema.validate_value(value=some_value)
[]

:class:`validate_value` returns an empty list if there are no errors.

The *name* parameter of the :meth:`validate_value`, is used to construct
friendly error messages. For example:

>>> meta.validate_value(prop_schema, None)
['The value for "some_prop" is required.']

The following example demonstrates how a :class:`PropertySchema` being
instantiated with a dictionary. Subsequently a bad value is passed with
multiple validation errors.

>>> other_schema = OnticProperty({
...     'name': 'other_prop',
...     'type': 'str',
...     'max': 3,
...     'enum': {'fish', 'dog', 'cat'}
... })
>>> ret = other_schema.validate_value('frog')
>>> assert len(ret) == 2
>>> assert ret[0] == '''The value "frog" for "other_prop" not in enumeration ['cat', 'dog', 'fish'].'''
>>> assert ret[1] == '''The value of "frog" for "other_prop" fails max of 3.'''

.. _property-schema-settings-table:

Available Property Schema Settings
-----------------------------------

The following table gives a listing of the property schema settings that can
be used to define properties. Details on the schema settings are provided
after the table.

.. table:: Property Schema Settings

    ============ ========= ======== ========  =================================
    Name         Type      Default  Required  Enumeration
    ============ ========= ======== ========  =================================
    type         str       None     False     bool, complex, date,
                 type                         datetime, dict, float, int, list,
                                              None, set, str, time
    default      None      None     False
    required     bool      False    False
    enum         set       None     False
    min          complex   None     False
                 date
                 datetime
                 float
                 int
                 time
    max          complex   None     False
                 date
                 datetime
                 float
                 int
                 time
    regex        str       None     False
    member_type  str       None     False     bool, complex, date,
                 type                         datetime, dict, float, int, list,
                                              None, set, str, time
    member_min   complex   None     False
                 date
                 datetime
                 float
                 int
                 time
    member_max   complex   None     False
                 date
                 datetime
                 float
                 int
                 time
    ============ ========= ======== ========  =================================

*type*
    The *type* settings restricts a property to a known type. If no type is
    defined, then any value type may be assigned to the property.

    The type definition can be by type or by string name. Both ``{type=int}``
    and ``{type='int'}`` are valid examples of type declaration.
*default*
    If no default is applied, then the default value will be ``None``. If a
    default value is supplied, it will only be applied under two conditions.
    A default value is applied during instantiation of an object of type
    :class:`PropertySchema`, :class:`~ontic.schema_type.SchemaType`,
    or :class:`~ontic.ontic_type.OnticType`. The other case is when an
    instance of on of the given types is perfected via the methods
    :func:`perfect_property_schema`, :func:`~ontic.schema_type.perfect_schema`,
    or :func:`~ontic.ontic_type.perfect_object`.

    The default is not applied during validation.

    For the collection types (dict, list, and set), the default value is deep
    copied. This is done to ensure that there is no sharing of collection
    instances or values.
*required*
    A *PropertySchema* with a required setting of *True*, will fail
    validation if the property value is *None*.
*enum*
    An *enum* setting is a set of values that the property value must adhere
    to. If the *type* setting is provided, then the choices provided by
    *enum* must be of that type. If no *type* is provided, then the choices
    in the *enum* set may be of any type, even mixed type.
*min*
    The *min* setting has differing behavior, based on the *type* setting. If
    no *type* setting is provided, then *min* test will not occur. For the
    boundable types (strings and collections) the *min* setting will test that
    the value length is not less than the minimum. For the comparable types
    (numeric and chronological) the *min* setting will test that the
    value is not less than the minimum.
*max*
    The *max setting has differing behavior, based on the *type* setting. If
    no *type* setting is provided, the *max* test will not occur. For the
    boundable types (strings and collections) the *max* setting will test that
    the value length is not more than the maximum. For the comparable types
    (numeric and chronological) the *max* setting will test that the
    value is not more than the maximum.
*regex*
    The *regex* setting is only tested if the *type* or *member_type* setting
    is 'str' and the *regex* setting is not None. When active, the *regex*
    setting will be used to test the given string value.  If the property
    value is 'None', then no regex testing will be done.
*member_type*
    The *member_type* setting is used to restrict the value type for property
    *type* 'list' or 'set'. It does so ensuring that each member of the
    collection is of the type designated by *member_type*.

    The type definition can be by type or by string name. Both
    ``{member_type=int}`` and ``{member_type='int'}`` are valid examples of
    type declaration.
*member_min*
    The *member_min* setting has differing behavior, based on the
    *member_type* setting. If no *member_type* setting is provided, then
    *member_min* test will not occur. For the boundable types
    (strings and collections), the *member_min* setting will test that the
    value length is not less than the minimum. For the comparable types
    (numeric and chronological) the *member_minimum* setting will test
    that the value is not less than the minimum.
*member_max*
    The *member_max* setting has differing behavior, based on the
    *member_max* setting. If no *member_type* setting is provided,
    then *member_max* test will not occur. For the boundable types
    (strings and collections), the *member_max* setting will test that the
    value length is not more than the maximum. For the comparable types
    (numeric and chronological) the *member_max* setting will test
    that the value is not more than the maximum.

"""
from typing import NoReturn, Any

from ontic import meta
from ontic import validation_exception


class OnticProperty(meta.Meta):
    """A class to define a schema for a property."""

    ONTIC_SCHEMA = meta.Meta({
        'name': meta.Meta({
            'name': 'name',
            'type': str,
            'default': None,
            'required': True,
            'enum': None,
            'min': 1,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'type': meta.Meta({
            'name': 'type',
            'type': (str, type),  # todo: raul - this could be restricted list
            'default': None,
            'required': False,
            'enum': meta.TYPE_SET + (None,),
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'default': meta.Meta({
            'name': 'default',
            'type': None,
            'default': None,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'required': meta.Meta({
            'name': 'required',
            'type': bool,
            'default': False,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'enum': meta.Meta({
            'name': 'enum',
            'type': set,
            'default': None,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'min': meta.Meta({
            'name': 'min',
            'type': tuple(meta.COMPARABLE_TYPES),
            'default': None,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'max': meta.Meta({
            'name': 'max',
            'type': tuple(meta.COMPARABLE_TYPES),
            'default': None,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'regex': meta.Meta({
            'name': 'regex',
            'type': str,
            'default': None,
            'required': False,
            'enum': None,
            'min': 1,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'member_type': meta.Meta({
            'name': 'member_type',
            'type': (str, type),  # todo: raul - this could be restricted list
            #  subclass testing.
            'default': None,
            'required': False,
            'enum': meta.TYPE_SET + (None,),
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'member_min': meta.Meta({
            'name': 'member_min',
            'type': tuple(meta.COMPARABLE_TYPES),
            'default': None,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
        'member_max': meta.Meta({
            'name': 'member_max',
            'type': tuple(meta.COMPARABLE_TYPES),
            'default': None,
            'required': False,
            'enum': None,
            'min': None,
            'max': None,
            'regex': None,
            'member_type': None,
            'member_min': None,
            'member_max': None,
        }),
    })

    def __init__(self, *args, **kwargs):
        super(OnticProperty, self).__init__(*args, **kwargs)

        self.perfect()
        self.validate()

    def perfect(self) -> NoReturn:
        """Method to ensure the completeness of a given schema property.

        :rtype: None
        :raises ValueError: If the candidate_property_type is None, or not
            of type *OnticProperty*.
        """
        perfect_property(self)

    def validate(
            self,
            raise_validation_exception: bool = True) -> (None, list[str]):
        """Method to validate a property schema definition.

        :param raise_validation_exception: If True, then
            *validate_property_type* will throw a *ValueException* upon
            validation failure. If False, then a list of validation errors is
            returned. Defaults to True.
        :return: If no validation errors are found, then *None* is returned.
            If validation fails, then a list of the errors is returned if the
            *raise_validation_exception* is set to True.
        :raises ValueError: *the_candidate_schema_property* is not an
            :class:`OnticProperty`.
        :raises ValidationException: A property of *candidate_property_type*
            does not meet schema requirements.
        """
        return validate_property(self, raise_validation_exception)

    def validate_value(self,
                       value: Any,
                       raise_validation_exception: bool = False) -> NoReturn:
        """Method that validates a value against the current schema."""
        value_errors = meta.validate_value(self, value)

        if value_errors and raise_validation_exception:
            raise validation_exception.ValidationException(value_errors)

        return value_errors


def perfect_property(ontic_property: 'OnticProperty') -> NoReturn:
    """Method to ensure the completeness of a given schema property.

    This method ensures completeness by stripping out any properties that
    are not defined by the schema definition. In addition, for any schema
    properties that are not included, the method will add those
    properties to the default value.

    :param ontic_property: The OnticProperty that is to be clean and restricted.
    :type ontic_property: :class:`OnticProperty`
    :rtype: None
    :raises ValueError: If the ontic_property is None, or not of type
        *OnticProperty*.
    """
    if ontic_property is None:
        raise ValueError('"ontic_property" must be provided.')
    if not isinstance(ontic_property, OnticProperty):
        raise ValueError('"ontic_property" must be OnticProperty type.')

    property_schema = ontic_property.get_schema()

    # remove un-necessary properties.
    extra_properties = set(ontic_property.keys()) - set(property_schema.keys())
    for property_name in extra_properties:
        del ontic_property[property_name]

    if 'type' in ontic_property:
        _perfect_type_setting(ontic_property)
    else:
        ontic_property.type = None

    if 'member_type' in ontic_property:
        _perfect_member_type_setting(ontic_property)
    else:
        ontic_property.member_type = None

    # set the default for the given property.
    for property_name, property_schema in property_schema.items():
        if property_name not in ontic_property:
            ontic_property[property_name] = None
        if ontic_property[property_name] is None:
            ontic_property[property_name] = property_schema.default


def validate_property(
        ontic_property: 'OnticProperty',
        raise_validation_exception: bool = True) -> (None, list[str]):
    """Method to validate a property schema definition.

    :param ontic_property: The schema property to be validated.
    :param raise_validation_exception: If True, then *validate_property_type*
        will throw a *ValueException* upon validation failure. If False,
        then a list of validation errors is returned. Defaults to True.
    :return: If no validation errors are found, then *None* is
        returned. If validation fails, then a list of the errors is returned
        if the *raise_validation_exception* is set to True.
    :raises ValueError: *the_candidate_schema_property* is not an
        :class:`~ontic.ontic_type.OnticType`.
    :raises ValidationException: A property of *ontic_property*
        does not meet schema requirements.
    """
    if ontic_property is None:
        raise ValueError('"ontic_property" must be provided.')
    if not isinstance(ontic_property, OnticProperty):
        raise ValueError('"ontic_property" must be OnticProperty type.')

    value_errors = []

    for property_schema in ontic_property.get_schema().values():
        property_value = ontic_property.get(property_schema.name, None)

        # todo: raul - for now skip validating compound schemas.
        if (isinstance(property_value, type) and
                issubclass(property_value, meta.Meta)):
            continue

        value_errors.extend(
            meta.validate_value(
                property_schema, property_value))

    if value_errors and raise_validation_exception:
        raise validation_exception.ValidationException(value_errors)

    return value_errors


def _perfect_type_setting(ontic_property: 'OnticProperty') -> None:
    """Perfect the type setting for a given candidate property schema."""
    if ontic_property.type is None:
        return

    candidate_type = ontic_property.type
    # coerce type declarations as string to base types.
    if isinstance(candidate_type, str):
        try:
            candidate_type = ontic_property.type = meta.TYPE_MAP[
                candidate_type]
        except KeyError:
            raise ValueError('Illegal type declaration: %s' %
                             ontic_property.type)

    # ensure that the type declaration is valid
    is_supported_type = candidate_type in meta.TYPE_SET
    is_meta_schema_type = issubclass(candidate_type, meta.Meta)
    if not (is_supported_type or is_meta_schema_type):
        raise ValueError('Illegal type declaration: %s' % str(candidate_type))


def _perfect_member_type_setting(ontic_property: 'OnticProperty') -> NoReturn:
    """Perfect the member_type setting for a given candidate property schema."""
    if ontic_property.member_type is None:
        return

    candidate_type = ontic_property.member_type
    # coerce type declarations as string to base types.
    if isinstance(candidate_type, str):
        try:
            candidate_type = ontic_property.member_type = meta.TYPE_MAP[
                candidate_type]
        except KeyError:
            raise ValueError('Illegal member_type declaration: %s' %
                             ontic_property.member_type)

    # ensure that the type declaration is valid
    is_supported_type = candidate_type in meta.TYPE_SET
    is_meta_schema_type = issubclass(candidate_type, meta.Meta)
    if not (is_supported_type or is_meta_schema_type):
        raise ValueError('Illegal type declaration: %s' % candidate_type)



"""The fundamental *Ontic* base data types for creation of derived child
classes.

.. image:: images/ontic_type.jpg

.. contents::

Usage
======

Create Ontic Types
--------------------

The *ontic_type* module provides the :class:`OnticType` and a set of
functions to handle the creation and validation of *OnticType* instances.

Construction of **Ontic** data types as a class definition::

    >>> class MyType(OnticType):
    ...     ONTIC_SCHEMA = Schema({
    ...         'some_property': {
    ...             'type': 'int',
    ...             'required': True,
    ...         },
    ...         'other_property': {
    ...             'type': 'str',
    ...             'required': False,
    ...             'enum': {'Enum1', 'Enum2', 'Enum3'}
    ...         },
    ...     })
    >>> my_object = MyType()
    >>> my_object.some_property = 7
    >>> # or
    >>> my_object['some_property'] = 7
    >>> validate_object(my_object)
    []

Dynamic Ontic Type Definition
-------------------------------

It is also possible to create :class:`OnticType` derived types dynamically
with the use of the :meth:`create_ontic_type` function.

    >>> some_type = create_ontic_type('SomeType', {'prop':{'type':'int'}})
    >>> my_object = some_type(prop=3)
    >>> my_object
    {'prop': 3}
    >>> my_object.prop
    3

"""
from copy import deepcopy
from typing import NoReturn

import ontic
from ontic import meta
from ontic.meta import Meta, COLLECTION_TYPES, TYPE_MAP
from ontic.schema import Schema
from ontic.validation_exception import ValidationException


class OnticType(Meta):
    """OnticType provides the **Ontic** schema interface.

    The **OnticType** provides the schema management functionality to a
    derived **Ontic** type instance.
    """

    def perfect(self) -> NoReturn:
        """Function to ensure complete attribute settings for a given object.

        Perfecting an object instance will strip out any properties not
        defined in
        the corresponding object type. If there are any missing properties in
        the
        object, those properties will be added and set to the default value or
        None, if no default has been set.

        For the collection types (dict, list, set), the default values are deep
        copied.

        :rtype: None
        """
        perfect_object(self)

    def validate(self,
                 raise_validation_exception: bool = True) -> (None, list[str]):
        """Validate the given OnticType instance against it's defined schema.

        :param raise_validation_exception: If True, then *validate_object* will
            throw a *ValueException* upon validation failure. If False, then a
            list of validation errors is returned. Defaults to True.
        :return: If no validation errors are found, then *None* is
            returned. If validation fails, then a list of the errors is returned
            if the *raise_validation_exception* is set to True.
        """
        return validate_object(self, raise_validation_exception)

    def validate_value(
            self,
            value_name: str,
            raise_validation_exception: bool = True) -> (None, list[str]):
        """Validate a target value of a given ontic object.

        :param value_name: The value name to validate.
        :param raise_validation_exception: If True, then *validate_object* will
            throw a *ValueException* upon validation failure. If False, then a
            list of validation errors is returned. Defaults to True.
        :return: If no validation errors are found, then *None* is
            returned. If validation fails, then a list of the errors is returned
            if the *raise_validation_exception* is set to True.
        """
        return validate_value(value_name, self, raise_validation_exception)


def create_ontic_type(name: str, schema: (dict, Schema)) -> OnticType:
    """Create an **Ontic** type to generate objects with a given schema.

    *create_ontic_type* function creates an :class:`OnticType` with a given
    name and schema definition. The schema definition can be a dict instance
    that is a valid  schema definition or a
    :class:`ontic.schema_type.SchemaType`. This makes the following forms
    valid::

        MyType = create_ontic_type('MyType', {'prop':{'type':'int'}})

        schema_instance = Schema(prop={'type':'int'})
        MyType = create_ontic_type('MyType', schema_instance)

    :param name: The name to apply to the created class, with
        :class:`OnticType` as parent.
    :param schema: A representation of the schema in dictionary format.
    :return: A class whose base is :class:`OnticType`.
    :raises ValueError: String name required. Dict or
        :class:`ontic.Schema` schema required.
    """
    if name is None or name == '':
        raise ValueError('The string "name" argument is required.')
    if schema is None:
        raise ValueError('The schema dictionary is required.')
    if not isinstance(schema, dict):
        raise ValueError('The schema must be a dict or SchemaType.')

    ontic_type = type(name, (OnticType,), dict())

    if not isinstance(schema, Schema):
        schema = Schema(schema)

    ontic_type.ONTIC_SCHEMA = schema

    return ontic_type


def perfect_object(the_object: OnticType) -> NoReturn:
    """Function to ensure complete attribute settings for a given object.

    Perfecting an object instance will strip out any properties not defined in
    the corresponding object type. If there are any missing properties in the
    object, those properties will be added and set to the default value or
    None, if no default has been set.

    For the collection types (dict, list, set), the default values are deep
    copied.

    :param the_object: Ab object instance that is to be perfected.
    """
    if the_object is None:
        raise ValueError('"the_object" must be provided.')
    if not isinstance(the_object, OnticType):
        raise ValueError('"the_object" must be OnticType type.')

    schema = the_object.get_schema()

    extra_properties = set(the_object.keys()) - set(schema.keys())
    for property_name in extra_properties:
        del the_object[property_name]

    for property_name, property_schema in schema.items():
        if property_name not in the_object:
            the_object[property_name] = None
        value = the_object[property_name]

        if value is None and property_schema.default is not None:
            if TYPE_MAP.get(property_schema.type) in COLLECTION_TYPES:
                value = the_object[property_name] = deepcopy(
                    property_schema.default)
            elif issubclass(property_schema.type, OnticType):
                value = the_object[property_name] = deepcopy(
                    property_schema.default)
            else:
                value = the_object[property_name] = property_schema.default

        if value is not None and isinstance(value, OnticType):
            value.perfect()


def validate_object(
        the_object: OnticType,
        raise_validation_exception: bool = True) -> (None, list[str]):
    """Function that will validate if an object meets the schema requirements.

    :param the_object: An object instant to be validity tested.
    :type the_object: :class:`OnticType`
    :param raise_validation_exception: If True, then *validate_object* will
        throw a *ValueException* upon validation failure. If False, then a
        list of validation errors is returned. Defaults to True.
    :return: If no validation errors are found, then *None* is
        returned. If validation fails, then a list of the errors is returned
        if the *raise_validation_exception* is set to True.
    :raises ValueError: If *the_object* is None or not of type
        :class:`~ontic.ontic_type.OnticType`.
    :raises ValidationException: A property of *the_object* does not meet
        schema requirements.
    """
    if not isinstance(the_object, OnticType):
        raise ValueError(
            'Validation can only support validation of objects derived from '
            'ontic.ontic_type.OnticType.')

    value_errors = []

    for property_name in the_object.get_schema().keys():
        value_errors.extend(validate_value(property_name, the_object, False))

    if value_errors and raise_validation_exception:
        raise ValidationException(value_errors)

    return value_errors


def validate_value(
        property_name: str,
        ontic_object: OnticType,
        raise_validation_exception: bool = False) -> (None, list[str]):
    """Validate a specific value of a given :class:`OnticType` instance.

    :param property_name: The value to be validated against the given
        **PropertyType**.
    :param ontic_object: Ontic defined object to be validated.
    :param raise_validation_exception: If True, then *validate_object* will
        throw a *ValueException* upon validation failure. If False, then a
        list of validation errors is returned. Defaults to True.
    :return: If no validation errors are found, then *None* is
        returned. If validation fails, then a list of the errors is returned
        if the *raise_validation_exception* is set to True.
    :raises ValueError: If *property_name* is not provided or is not a valid
        string.
    :raises ValueError: If *ontic_object* is None, or not instance of
        *OnticType*.
    :raises ValidationException: If the validation is not successful. The
        *ValidationException* will not be raised if
        *raise_validation_exception* is
        set to False.
    """
    if property_name is None:
        raise ValueError(
            '"property_name" is required, cannot be None.')
    if not isinstance(property_name, str) or len(property_name) < 1:
        raise ValueError('"property_name" is not a valid string.')
    if ontic_object is None:
        raise ValueError(
            '"ontic_object" is required, cannot be None.')
    if not isinstance(ontic_object, OnticType):
        raise ValueError(
            '"ontic_object" must be OnticType or child type of OnticType.')

    value_errors = []

    property_schema = ontic_object.get_schema().get(property_name)
    if property_schema is None:
        raise ValueError(
            '"%s" is not a recognized property.' % property_name)

    value = ontic_object.get(property_name, None)

    value_errors.extend(
        meta.validate_value(property_schema, value))

    # if a value is an OnticType, then have it self validate.
    is_ontic_type = (property_schema.type is not None and
                     issubclass(property_schema.type, OnticType))
    child_errors = None
    if is_ontic_type and value is not None:
        child_errors = value.validate(raise_validation_exception=False)
    if child_errors:
        error_msg = 'The child property %s, has errors:: %s' % (
            property_name, ' || '.join(child_errors))
        value_errors.append(error_msg)

    if value_errors and raise_validation_exception:
        raise ValidationException(value_errors)

    return value_errors



"""The root class for the construction of *Ontic* implementation types.

.. image:: images/ontic_core.jpg

.. contents::

"""

from copy import copy, deepcopy


class Core(dict):
    """The root type of *Ontic* types.
    
    **Core** provides for *Ontic* object properties to be accessible as
    either dict key-value pairs or as object attributes. Core also supports
    dict style initialization.

    Dict Style Initialization
        Core() -> new empty Core

        Core(mapping) -> new Core initialized from a mapping
        object's (key, value) pairs

        Core(iterable) -> new Core initialized as if via::

            d = Core()
            for k, v in iterable:
                d[k] = v

        Core(**kwargs) -> new Core initialized with the
        name=value pairs in the keyword argument list.  For example::

            Core(one=1, two=2)
    
    Example dict style and object style initialization::

    >>> some_object = Core({'key1': 'value1'}) # Dict style initialization
    >>> other_object = Core(key1='value1') # Object style initialization

    Example dict style and object style property access::

    >>> # Object attribute access to value
    >>> assert some_object.key1 == 'value1'
    >>> assert other_object.key1 == 'value1'
    >>> # Dict key access to value
    >>> assert some_object['key1'] == 'value1'
    >>> assert other_object['key1'] == 'value1'
    >>> # Dynamic property assignment is supported
    >>> some_object.key2 = 'value2' # Object value initialization
    >>> assert some_object['key2'] == 'value2'
    >>> other_object['key3'] = 'value3' # Dict style key-value assignment
    >>> assert other_object.key3 == 'value3'
    """

    def __init__(self, *args, **kwargs):
        super(Core, self).__init__(*args, **kwargs)

        self.__dict__ = self

    def __copy__(self) -> 'Core':
        return type(self)(copy(dict(self)))

    def __deepcopy__(self, memo) -> 'Core':
        the_copy = dict(self.__dict__)
        return type(self)(deepcopy(the_copy))



"""The setuptools setup file."""
from setuptools import setup

# todo: raul - if the setup barfs on README.txt the use the setup.cfg file.as
# the tag is [metadata]/n description-file = README.md

with open('README.md') as file:
    long_description = file.read()

with open('VERSION') as version_file:
    version = version_file.read().strip()

setup(
    name='ontic',
    version=version,
    author='Raul Gonzalez',
    author_email='mindbender@gmail.com',
    url='https://github.com/neoinsanity/ontic',
    license='Apache License 2.0',
    description='Qualities and Quantities are the stuff of Objects.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['ontic', ],
    install_requires=[],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
    ]
)
